// ------------------------------ elo.c ------------------------------
/*
    LibCapy - a general purpose library of C functions and data structures
    Copyright (C) 2021-2024 Pascal Baillehache baillehache.pascal@gmail.com
    https://baillehachepascal.dev
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
#include "elo.h"

// 'that' keyword
#define CapyThatElo CapyElo* that = (CapyElo*)capyThat

// Get the Elo delta given two Elos and a result. This delta is to be
// added to the first Elo and substracted to the second.
// Input:
//   eloA: first Elo
//   result: result of eloA against eloB (e.g., if result is 'win' it
//           means that eloA has won over eloB)
//   eloB: second Elo
// Output:
//   Return the Elo delta to be added to eloA and substracted to eloB
static double Get(
         double const eloA,
  CapyEloResult const result,
         double const eloB) {
  CapyThatElo;

  // Calculate the Elo delta
  double expRes = 1.0 / (1.0 + pow(10.0, (eloB - eloA) / 400.0));
  double eloDelta = that->k * (((double)result) * 0.5 - expRes);

  // Return the Elo delta
  return eloDelta;
}

// Free the memory used by a CapyElo
static void Destruct(void) {
  return;
}

// Create a CapyElo
// Output:
//   Return a CapyElo
CapyElo CapyEloCreate(void) {
  return (CapyElo){
    .k = CAPY_ELO_K,
    .destruct = Destruct,
    .get = Get,
  };
}

// Allocate memory for a new CapyElo and create it
// Output:
//   Return a CapyElo
// Exception:
//   May raise CapyExc_MallocFailed.
CapyElo* CapyEloAlloc(void) {
  CapyElo* that = NULL;
  safeMalloc(that, 1);
  if(!that) return NULL;
  *that = CapyEloCreate();
  return that;
}

// Free the memory used by a CapyElo* and reset '*that' to NULL
// Input:
//   that: a pointer to the CapyElo to free
void CapyEloFree(CapyElo** const that) {
  if(that == NULL || *that == NULL) return;
  $(*that, destruct)();
  free(*that);
  *that = NULL;
}
