// ---------------------- frequentistHypothesisTesting.h ---------------------
/*
    LibCapy - a general purpose library of C functions and data structures
    Copyright (C) 2021-2024 Pascal Baillehache baillehache.pascal@gmail.com
    https://baillehachepascal.dev
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef CAPY_FREQUENTIST_HYPOTHESIS_METHOD_H
#define CAPY_FREQUENTIST_HYPOTHESIS_METHOD_H
#include "externalHeaders.h"
#include "cext.h"
#include "distribution.h"

// Description:
// Class implementing the frequentist hypothesis testing method.

// Frequentist hypothesis testing class.
typedef struct CapyFHT {

  // Confidence that an event matches the null hypothesis (in [0,1])
  double statisticalPower;

  // Confidence that an event doesn't match the alternate hypothesis (in [0,1])
  double pValue;

  // Distribution of the null hypothesis
  CapyDist const* nullHypo;

  // Distribution of the alternate hypothesis
  CapyDist const* altHypo;

  // Destructor
  void (*destruct)(void);

  // Check if an event matches the null hypothesis
  // Input:
  //   evt: the event
  // Output:
  //   Return true if the event matches, else false.
  bool (*isEvtMatchingNullHypo)(CapyDistEvt const* const evt);
} CapyFHT;

// Create a CapyFHT
// Input:
//   statisticalPower: confidence that an event doesn't match the alternate
//                     hypothesis (in [0,1])
//             pValue: confidence that an event matches the null hypothesis
//                     (in [0,1])
//           nullHypo: distribution of the null hypothesis
//             altHyp: distribution of the alternate hypothesis
// Output:
//   Return a CapyFHT
CapyFHT CapyFHTCreate(
           double const statisticalPower,
           double const pValue,
  CapyDist const* const nullHypo,
  CapyDist const* const altHypo);

// Allocate memory for a new CapyFHT and create it.
// Input:
//   statisticalPower: confidence that an event doesn't match the alternate
//                     hypothesis (in [0,1])
//             pValue: confidence that an event matches the null hypothesis
//                     (in [0,1])
//           nullHypo: distribution of the null hypothesis
//             altHyp: distribution of the alternate hypothesis
// Output:
//   Return a CapyFHT
// Exception:
//   May raise CapyExc_MallocFailed.
CapyFHT* CapyFHTAlloc(
           double const statisticalPower,
           double const pValue,
  CapyDist const* const nullHypo,
  CapyDist const* const altHypo);

// Free the memory used by a CapyFHT* and reset '*that' to NULL
// Input:
//   that: a pointer to the CapyFHT to free
void CapyFHTFree(CapyFHT** const that);
#endif
