// ---------------------------- lsystem.c ---------------------------
/*
    LibCapy - a general purpose library of C functions and data structures
    Copyright (C) 2021-2024 Pascal Baillehache baillehache.pascal@gmail.com
    https://baillehachepascal.dev
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
#include "lsystem.h"

// Rule for Lindenmayer's algae L-System
static CapyLSysLindenMayerAlgaeState* LindenmayerAlgaeRule(
   size_t const idx,
  size_t* const lenOut) {
  CapyLSysLindenmayerAlgae* that = (CapyLSysLindenmayerAlgae*)capyThat;
  CapyLSysLindenMayerAlgaeState* out = NULL;
  if(that->state[idx].val == 'A') {
    safeMalloc(out, 2);
    assert(out);
    out[0] = (CapyLSysLindenMayerAlgaeState){.val='A'};
    out[1] = (CapyLSysLindenMayerAlgaeState){.val='B'};
    *lenOut = 2;
  } else if(that->state[idx].val == 'B') {
    safeMalloc(out, 1);
    assert(out);
    out[0] = (CapyLSysLindenMayerAlgaeState){.val='A'};
    *lenOut = 1;
  }
  return out;
}

// Lindenmayer's algae L-System definition
CapyDefLSystem(
  CapyLSysLindenmayerAlgae, CapyLSysLindenMayerAlgaeState, \
  LindenmayerAlgaeRule)

// Rule for Koch snowlfake L-System
static CapyLSysKochSnowflakeState* KochSnowflakeRule(
   size_t const idx,
  size_t* const lenOut) {
  CapyLSysKochSnowflake* that = (CapyLSysKochSnowflake*)capyThat;
  CapyLSysKochSnowflakeState* out = NULL;
  if(that->state[idx].val == 'F') {
    safeMalloc(out, 10);
    assert(out);
    out[0] = (CapyLSysKochSnowflakeState){.val='D'};
    out[1] = (CapyLSysKochSnowflakeState){.val='F'};
    out[2] = (CapyLSysKochSnowflakeState){.val='+'};
    out[3] = (CapyLSysKochSnowflakeState){.val='F'};
    out[4] = (CapyLSysKochSnowflakeState){.val='-'};
    out[5] = (CapyLSysKochSnowflakeState){.val='-'};
    out[6] = (CapyLSysKochSnowflakeState){.val='F'};
    out[7] = (CapyLSysKochSnowflakeState){.val='+'};
    out[8] = (CapyLSysKochSnowflakeState){.val='F'};
    out[9] = (CapyLSysKochSnowflakeState){.val='U'};
    *lenOut = 10;
  } else {
    safeMalloc(out, 1);
    assert(out);
    out[0] = (CapyLSysKochSnowflakeState){.val=that->state[idx].val};
    *lenOut = 1;
  }
  return out;
}

// Koch snowflake L-System definition
CapyDefLSystem(
  CapyLSysKochSnowflake, CapyLSysKochSnowflakeState, KochSnowflakeRule)

// Rule for fractal binary tree L-System
static CapyLSysFractalBinaryTreeState* FractalBinaryTreeRule(
   size_t const idx,
  size_t* const lenOut) {
  CapyLSysFractalBinaryTree* that = (CapyLSysFractalBinaryTree*)capyThat;
  CapyLSysFractalBinaryTreeState* out = NULL;
  if(that->state[idx].val == '0') {
    safeMalloc(out, 5);
    assert(out);
    out[0] = (CapyLSysFractalBinaryTreeState){.val='1'};
    out[1] = (CapyLSysFractalBinaryTreeState){.val='['};
    out[2] = (CapyLSysFractalBinaryTreeState){.val='0'};
    out[3] = (CapyLSysFractalBinaryTreeState){.val=']'};
    out[4] = (CapyLSysFractalBinaryTreeState){.val='0'};
    *lenOut = 5;
  } else if(that->state[idx].val == '1') {
    safeMalloc(out, 2);
    assert(out);
    out[0] = (CapyLSysFractalBinaryTreeState){.val='1'};
    out[1] = (CapyLSysFractalBinaryTreeState){.val='1'};
    *lenOut = 2;
  } else {
    safeMalloc(out, 1);
    assert(out);
    out[0] = (CapyLSysFractalBinaryTreeState){.val=that->state[idx].val};
    *lenOut = 1;
  }
  return out;
}

// Fractal binary tree L-System definition
CapyDefLSystem(
  CapyLSysFractalBinaryTree, CapyLSysFractalBinaryTreeState,
  FractalBinaryTreeRule)
