// ---------------------------- rungekutta.c ---------------------------
/*
    LibCapy - a general purpose library of C functions and data structures
    Copyright (C) 2021-2024 Pascal Baillehache info@baillehachepascal.dev
    https://baillehachepascal.dev
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
#include "rungekutta.h"

// 'that' keyword
#define CapyThatRungeKutta CapyRungeKutta* that = (CapyRungeKutta*)capyThat

// Approximate x(t) using RK4
// Input:
//   t: value for which we want to approximate x(t)
// Output:
//   Return the approximated value of x(t). 't' is expected to be greater than
//   't0', else return 'x0'. 't-t0' is expected to be a divisible by 'deltaT',
//   else return the value for the nearest divisible value not greater than
//   't'.
static double Eval(double const t) {
  CapyThatRungeKutta;
  double in[2] = {that->initVals[0], that->initVals[1]};
  while(in[1] < t) {
    $(that, step)(in);
  }
  return in[0];
}

// Approximate x(t) using RK4 on an interval
// Input:
//   t: approximate x(t) on the range [t0, t]
// Output:
//   Return the approximated values of x(t). 't' is expected to be greater
//   than 't0', else return on single value equal to 'x0'. 't-t0' is expected
//   to be a divisible by 'deltaT', else use the nearest divisible value not
//   greater than 't' instead.
static CapyVec EvalRange(double const t) {
  CapyThatRungeKutta;
  size_t const nbEval = 1 + (size_t)((t - that->initVals[1]) / that->deltaT);
  CapyVec res = CapyVecCreate(nbEval);
  res.vals[0] = that->initVals[0];
  size_t iEval = 1;
  double in[2] = {that->initVals[0], that->initVals[1]};
  while(in[1] < t) {
    $(that, step)(in);
    res.vals[iEval] = in[0];
    iEval += 1;
  }
  return res;
}

// Run one step
// Input:
//   vals: current values [x, t]
// Output:
//   'vals' is updated with the result of one step of RK4
static void Step(double* const vals) {
  CapyThatRungeKutta;
  double k[4] = {0.0, 0.0, 0.0, 0.0};
  $(that->derivative, eval)(vals, k + 0);
  double in1[2] = {
    vals[0] + that->deltaT * 0.5 * k[0],
    vals[1] + that->deltaT * 0.5
  };
  $(that->derivative, eval)(in1, k + 1);
  double in2[2] = {
    vals[0] + that->deltaT * 0.5 * k[1],
    vals[1] + that->deltaT * 0.5
  };
  $(that->derivative, eval)(in2, k + 2);
  double in3[2] = {
    vals[0] + that->deltaT * k[2],
    vals[1] + that->deltaT
  };
  $(that->derivative, eval)(in3, k + 3);
  vals[0] +=
    that->deltaT / 6.0 * (k[0] + 2.0 * k[1] + 2.0 * k[2] + k[3]);
  vals[1] += that->deltaT;
}

// Free the memory used by a CapyRungeKutta
static void Destruct(void) {
  CapyThatRungeKutta;
  *that = (CapyRungeKutta){0};
}

// Create a CapyRungeKutta
// Input:
//   derivative: the derivative function
// Output:
//   Return a CapyRungeKutta
CapyRungeKutta CapyRungeKuttaCreate(CapyMathFun* const derivative) {
  CapyRungeKutta that = {
    .derivative = derivative,
    .initVals = {0.0, 0.0},
    .deltaT = 1e-3,
    .eval = Eval,
    .evalRange = EvalRange,
    .step = Step,
    .destruct = Destruct,
  };
  if(derivative->dimIn != 2 || derivative->dimOut != 1) {
    raiseExc(CapyExc_InvalidParameters);
  }
  return that;
}

// Allocate memory for a new CapyRungeKutta and create it
// Input:
//   derivative: the derivative function
// Output:
//   Return a CapyRungeKutta
// Exception:
//   May raise CapyExc_MallocFailed.
CapyRungeKutta* CapyRungeKuttaAlloc(CapyMathFun* const derivative) {
  CapyRungeKutta* that = NULL;
  safeMalloc(that, 1);
  if(!that) return NULL;
  *that = CapyRungeKuttaCreate(derivative);
  return that;
}

// Free the memory used by a CapyRungeKutta* and reset '*that' to NULL
// Input:
//   that: a pointer to the CapyRungeKutta to free
void CapyRungeKuttaFree(CapyRungeKutta** const that) {
  if(that == NULL || *that == NULL) return;
  $(*that, destruct)();
  free(*that);
  *that = NULL;
}
