// ---------------------------------- rungekutta.h ---------------------------------
/*
    LibCapy - a general purpose library of C functions and data structures
    Copyright (C) 2021-2024 Pascal Baillehache info@baillehachepascal.dev
    https://baillehachepascal.dev
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef CAPY_RUNGEKUTTA_H
#define CAPY_RUNGEKUTTA_H
#include "externalHeaders.h"
#include "cext.h"
#include "mathfun.h"

// Description:
// RungeKutta implementation.

// RungeKutta object
typedef struct CapyRungeKutta {

  // Derivative function F(), such as we want to solve x(t) knowing x(t0) and
  // F(x,t)=dx/dt. Input dimension of F is 2, output dimension is 1.
  CapyMathFun* derivative;

  // Initial values [x0, t0] (default: [0, 0])
  double initVals[2];

  // Discretise step delta t (default: 1e-3)
  double deltaT;

  // Approximate x(t) using RK4
  // Input:
  //   t: value for which we want to approximate x(t)
  // Output:
  //   Return the approximated value of x(t). 't' is expected to be greater than
  //   't0', else return 'x0'. 't-t0' is expected to be a divisible by 'deltaT',
  //   else return the value for the nearest divisible value not greater than
  //   't'.
  double (*eval)(double const t);

  // Approximate x(t) using RK4 on an interval
  // Input:
  //   t: approximate x(t) on the range [t0, t]
  // Output:
  //   Return the approximated values of x(t). 't' is expected to be greater
  //   than 't0', else return on single value equal to 'x0'. 't-t0' is expected
  //   to be a divisible by 'deltaT', else use the nearest divisible value not
  //   greater than 't' instead.
  CapyVec (*evalRange)(double const t);

  // Run one step
  // Input:
  //   vals: current values [x, t]
  // Output:
  //   'vals' is updated with the result of one step of RK4
  void (*step)(double* const vals);

  // Destructor
  void (*destruct)(void);
} CapyRungeKutta;

// Create a CapyRungeKutta
// Input:
//   derivative: the derivative function
// Output:
//   Return a CapyRungeKutta
CapyRungeKutta CapyRungeKuttaCreate(CapyMathFun* const derivative);

// Allocate memory for a new CapyRungeKutta and create it
// Input:
//   derivative: the derivative function
// Output:
//   Return a CapyRungeKutta
// Exception:
//   May raise CapyExc_MallocFailed.
CapyRungeKutta* CapyRungeKuttaAlloc(CapyMathFun* const derivative);

// Free the memory used by a CapyRungeKutta* and reset '*that' to NULL
// Input:
//   that: a pointer to the CapyRungeKutta to free
void CapyRungeKuttaFree(CapyRungeKutta** const that);
#endif
